@@ -896,12 +896,12 @@ connect_enable_indications_ipv4_ready (QmiClientWds *client,
 
     ctx->event_report_ipv4_indication_id =
         connect_enable_indications_ready (client, res, ctx->self, &ctx->error_ipv4);
-    if (!ctx->event_report_ipv4_indication_id) {
-        ctx->step = CONNECT_STEP_LAST;
-        return;
-    }
 
-    ctx->step++;
+    if (!ctx->event_report_ipv4_indication_id)
+        ctx->step = CONNECT_STEP_LAST;
+    else
+        ctx->step++;
+
     connect_context_step (ctx);
 }
 
@@ -914,12 +914,12 @@ connect_enable_indications_ipv6_ready (QmiClientWds *client,
 
     ctx->event_report_ipv6_indication_id =
         connect_enable_indications_ready (client, res, ctx->self, &ctx->error_ipv6);
-    if (!ctx->event_report_ipv6_indication_id) {
-        ctx->step = CONNECT_STEP_LAST;
-        return;
-    }
 
-    ctx->step++;
+    if (!ctx->event_report_ipv6_indication_id)
+        ctx->step = CONNECT_STEP_LAST;
+    else
+        ctx->step++;
+
     connect_context_step (ctx);
 }
 
@@ -1156,8 +1156,11 @@ connect_context_step (ConnectContext *ctx)
                                                ctx->cancellable,
                                                (GAsyncReadyCallback) connect_enable_indications_ipv4_ready,
                                                ctx);
-#endif
         return;
+#else
+        /* Just fall down */
+        ctx->step++;
+#endif
 
     case CONNECT_STEP_START_NETWORK_IPV4: {
         QmiMessageWdsStartNetworkInput *input;
@@ -1259,8 +1262,11 @@ connect_context_step (ConnectContext *ctx)
                                                ctx->cancellable,
                                                (GAsyncReadyCallback) connect_enable_indications_ipv6_ready,
                                                ctx);
-#endif
         return;
+#else
+        /* Just fall down */
+        ctx->step++;
+#endif
 
     case CONNECT_STEP_START_NETWORK_IPV6: {
         QmiMessageWdsStartNetworkInput *input;
