@@ -214,9 +214,20 @@ apply_pre_probing_filters (MMPlugin *self,
     }
 
     /* The plugin may specify that only some drivers are supported, or that some
-     * drivers are not supported. If that is the case, filter by driver */
+     * drivers are not supported. If that is the case, filter by driver.
+     *
+     * The QMI and MBIM *forbidden* drivers filter is implicit. This is, if the
+     * plugin doesn't explicitly specify that QMI is allowed and we find a QMI
+     * port, the plugin will filter the device. Same for MBIM.
+     *
+     * The opposite, though, is not applicable. If the plugin specifies that QMI
+     * is allowed, we won't take that as a mandatory requirement to look for the
+     * QMI driver (as the plugin may handle non-QMI modems as well)
+     */
     if (self->priv->drivers ||
-        self->priv->forbidden_drivers) {
+        self->priv->forbidden_drivers ||
+        !self->priv->qmi ||
+        !self->priv->mbim) {
         static const gchar *virtual_drivers [] = { "virtual", NULL };
         const gchar **drivers;
 
@@ -254,8 +265,9 @@ apply_pre_probing_filters (MMPlugin *self,
                 return TRUE;
             }
         }
+
         /* Filtering by forbidden drivers */
-        else {
+        if (self->priv->forbidden_drivers) {
             for (i = 0; self->priv->forbidden_drivers[i]; i++) {
                 guint j;
 
@@ -270,6 +282,36 @@ apply_pre_probing_filters (MMPlugin *self,
                 }
             }
         }
+
+        /* Implicit filter for forbidden QMI driver */
+        if (!self->priv->qmi) {
+            guint j;
+
+            for (j = 0; drivers[j]; j++) {
+                /* If we match the QMI driver: unsupported */
+                if (g_str_equal (drivers[j], "qmi_wwan")) {
+                    mm_dbg ("(%s) [%s] filtered by implicit QMI driver",
+                            self->priv->name,
+                            g_udev_device_get_name (port));
+                    return TRUE;
+                }
+            }
+        }
+
+        /* Implicit filter for forbidden MBIM driver */
+        if (!self->priv->mbim) {
+            guint j;
+
+            for (j = 0; drivers[j]; j++) {
+                /* If we match the MBIM driver: unsupported */
+                if (g_str_equal (drivers[j], "cdc_mbim")) {
+                    mm_dbg ("(%s) [%s] filtered by implicit MBIM driver",
+                            self->priv->name,
+                            g_udev_device_get_name (port));
+                    return TRUE;
+                }
+            }
+        }
     }
 
     vendor = mm_device_get_vendor (device);
@@ -308,6 +350,14 @@ apply_pre_probing_filters (MMPlugin *self,
             if (!self->priv->product_ids[i].l)
                 product_filtered = TRUE;
         }
+
+        /* When both vendor ids and product ids are given, it may be the case that
+         * we're allowing a full VID1 and only a subset of another VID2, so try to
+         * handle that properly. */
+        if (vendor_filtered && !product_filtered)
+            vendor_filtered = FALSE;
+        if (product_filtered && self->priv->vendor_ids && !vendor_filtered)
+            product_filtered = FALSE;
     }
 
     /* If we got filtered by vendor or product IDs; mark it as unsupported only if:
@@ -486,8 +536,8 @@ apply_post_probing_filters (MMPlugin *self,
                 gchar *casefolded_vendor;
                 gchar *casefolded_product;
 
-                casefolded_vendor = g_utf8_casefold (self->priv->product_strings[i].l, -1);
-                casefolded_product = g_utf8_casefold (self->priv->product_strings[i].r, -1);
+                casefolded_vendor = g_utf8_casefold (self->priv->forbidden_product_strings[i].l, -1);
+                casefolded_product = g_utf8_casefold (self->priv->forbidden_product_strings[i].r, -1);
                 found = (!!strstr (vendor, casefolded_vendor) &&
                          !!strstr (product, casefolded_product));
                 g_free (casefolded_vendor);
@@ -532,125 +582,119 @@ apply_post_probing_filters (MMPlugin *self,
 
 /* Context for the asynchronous probing operation */
 typedef struct {
-    GSimpleAsyncResult *result;
     MMPlugin *self;
-    MMPortProbeFlag flags;
     MMDevice *device;
+    MMPortProbeFlag flags;
 } PortProbeRunContext;
 
+static void
+port_probe_run_context_free (PortProbeRunContext *ctx)
+{
+    g_object_unref (ctx->device);
+    g_object_unref (ctx->self);
+    g_slice_free (PortProbeRunContext, ctx);
+}
+
 static void
 port_probe_run_ready (MMPortProbe *probe,
                       GAsyncResult *probe_result,
-                      PortProbeRunContext *ctx)
+                      GTask *task)
 {
     GError *error = NULL;
+    MMPluginSupportsResult result = MM_PLUGIN_SUPPORTS_PORT_UNKNOWN;
+    PortProbeRunContext *ctx;
 
     if (!mm_port_probe_run_finish (probe, probe_result, &error)) {
         /* Probing failed saying the port is unsupported. This is not to be
          * treated as a generic error, the plugin is just telling us as nicely
          * as it can that the port is not supported, so don't warn these cases.
          */
-        if (g_error_matches (error,
-                             MM_CORE_ERROR,
-                             MM_CORE_ERROR_UNSUPPORTED)) {
-            g_simple_async_result_set_op_res_gpointer (ctx->result,
-                                                       GUINT_TO_POINTER (MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED),
-                                                       NULL);
-            g_error_free (error);
+        if (g_error_matches (error, MM_CORE_ERROR, MM_CORE_ERROR_UNSUPPORTED)) {
+            result = MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED;
+            goto out;
         }
+
         /* Probing failed but the plugin tells us to retry; so we'll defer the
          * probing a bit */
-        else if (g_error_matches (error,
-                                  MM_CORE_ERROR,
-                                  MM_CORE_ERROR_RETRY)) {
-            g_simple_async_result_set_op_res_gpointer (ctx->result,
-                                                       GUINT_TO_POINTER (MM_PLUGIN_SUPPORTS_PORT_DEFER),
-                                                       NULL);
-            g_error_free (error);
+        if (g_error_matches (error, MM_CORE_ERROR, MM_CORE_ERROR_RETRY)) {
+            result = MM_PLUGIN_SUPPORTS_PORT_DEFER;
+            goto out;
         }
+
         /* For remaining errors, just propagate them */
-        else {
-            g_simple_async_result_take_error (ctx->result, error);
-        }
-    } else {
-        /* Probing succeeded */
-        MMPluginSupportsResult supports_result;
-
-        if (!apply_post_probing_filters (ctx->self, ctx->flags, probe)) {
-            /* Port is supported! */
-            supports_result = MM_PLUGIN_SUPPORTS_PORT_SUPPORTED;
-
-            /* If we were looking for AT ports, and the port is AT,
-             * and we were told that only one AT port is expected, cancel AT
-             * probings in the other available support tasks of the SAME
-             * device. */
-            if (ctx->self->priv->single_at &&
-                ctx->flags & MM_PORT_PROBE_AT &&
-                mm_port_probe_is_at (probe)) {
-                GList *l;
-
-                for (l = mm_device_peek_port_probe_list (ctx->device); l; l = g_list_next (l)) {
-                    if (l->data != probe) {
-                        mm_port_probe_run_cancel_at_probing (MM_PORT_PROBE (l->data));
-                    }
-                }
-            }
-        } else {
-            /* Unsupported port, remove from internal tracking HT */
-            supports_result = MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED;
-        }
-
-        g_simple_async_result_set_op_res_gpointer (ctx->result,
-                                                   GUINT_TO_POINTER (supports_result),
-                                                   NULL);
+        g_task_return_error (task, error);
+        g_object_unref (task);
+        return;
     }
 
-    /* Complete the async supports port request */
-    g_simple_async_result_complete_in_idle (ctx->result);
+    /* Probing succeeded, recover context */
+    ctx = g_task_get_task_data (task);
 
-    g_object_unref (ctx->device);
-    g_object_unref (ctx->result);
-    g_object_unref (ctx->self);
-    g_free (ctx);
+    /* Apply post probing filters */
+    if (!apply_post_probing_filters (ctx->self, ctx->flags, probe)) {
+        /* Port is supported! */
+        result = MM_PLUGIN_SUPPORTS_PORT_SUPPORTED;
+
+        /* If we were looking for AT ports, and the port is AT,
+         * and we were told that only one AT port is expected, cancel AT
+         * probings in the other available support tasks of the SAME
+         * device. */
+        if (ctx->self->priv->single_at &&
+            ctx->flags & MM_PORT_PROBE_AT &&
+            mm_port_probe_is_at (probe)) {
+            GList *l;
+
+            for (l = mm_device_peek_port_probe_list (ctx->device); l; l = g_list_next (l)) {
+                if (l->data != probe)
+                    mm_port_probe_run_cancel_at_probing (MM_PORT_PROBE (l->data));
+            }
+        }
+    } else
+        /* Filtered by post probing filters */
+        result = MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED;
+
+out:
+    /* Complete action */
+    g_clear_error (&error);
+    g_task_return_int (task, result);
+    g_object_unref (task);
 }
 
+G_STATIC_ASSERT (MM_PLUGIN_SUPPORTS_PORT_UNKNOWN == -1);
+
 MMPluginSupportsResult
-mm_plugin_supports_port_finish (MMPlugin *self,
-                                GAsyncResult *result,
-                                GError **error)
+mm_plugin_supports_port_finish (MMPlugin      *self,
+                                GAsyncResult  *result,
+                                GError       **error)
 {
-    g_return_val_if_fail (MM_IS_PLUGIN (self),
-                          MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED);
-    g_return_val_if_fail (G_IS_ASYNC_RESULT (result),
-                          MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED);
+    g_return_val_if_fail (MM_IS_PLUGIN (self), MM_PLUGIN_SUPPORTS_PORT_UNKNOWN);
+    g_return_val_if_fail (G_IS_TASK (result), MM_PLUGIN_SUPPORTS_PORT_UNKNOWN);
 
-    /* Propagate error, if any */
-    if (g_simple_async_result_propagate_error (G_SIMPLE_ASYNC_RESULT (result), error)) {
-        return MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED;
-    }
-
-    return (MMPluginSupportsResult) GPOINTER_TO_UINT (g_simple_async_result_get_op_res_gpointer (G_SIMPLE_ASYNC_RESULT (result)));
+    return (MMPluginSupportsResult) g_task_propagate_int (G_TASK (result), error);
 }
 
 void
-mm_plugin_supports_port (MMPlugin *self,
-                         MMDevice *device,
-                         GUdevDevice *port,
-                         GAsyncReadyCallback callback,
-                         gpointer user_data)
+mm_plugin_supports_port (MMPlugin            *self,
+                         MMDevice            *device,
+                         GUdevDevice         *port,
+                         GCancellable        *cancellable,
+                         GAsyncReadyCallback  callback,
+                         gpointer             user_data)
 {
     MMPortProbe *probe = NULL;
-    GSimpleAsyncResult *async_result;
+    GTask *task;
     PortProbeRunContext *ctx;
     gboolean need_vendor_probing;
     gboolean need_product_probing;
     MMPortProbeFlag probe_run_flags;
     gchar *probe_list_str;
 
-    async_result = g_simple_async_result_new (G_OBJECT (self),
-                                              callback,
-                                              user_data,
-                                              mm_plugin_supports_port);
+    g_return_if_fail (MM_IS_PLUGIN (self));
+    g_return_if_fail (MM_IS_DEVICE (device));
+    g_return_if_fail (G_UDEV_IS_DEVICE (port));
+
+    /* Create new cancellable task */
+    task = g_task_new (self, cancellable, callback, user_data);
 
     /* Apply filters before launching the probing */
     if (apply_pre_probing_filters (self,
@@ -659,46 +703,40 @@ mm_plugin_supports_port (MMPlugin *self,
                                    &need_vendor_probing,
                                    &need_product_probing)) {
         /* Filtered! */
-        g_simple_async_result_set_op_res_gpointer (async_result,
-                                                   GUINT_TO_POINTER (MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED),
-                                                   NULL);
-        g_simple_async_result_complete_in_idle (async_result);
-        goto out;
+        g_task_return_int (task, MM_PLUGIN_SUPPORTS_PORT_UNSUPPORTED);
+        g_object_unref (task);
+        return;
     }
 
     /* Need to launch new probing */
-    probe = MM_PORT_PROBE (mm_device_get_port_probe (device, port));
+    probe = MM_PORT_PROBE (mm_device_peek_port_probe (device, port));
     if (!probe) {
         /* This may happen if the ports get removed from the device while
          * probing is ongoing */
-        g_simple_async_result_set_error (async_result,
-                                         MM_CORE_ERROR,
-                                         MM_CORE_ERROR_FAILED,
-                                         "(%s) Missing port probe for port (%s/%s)",
-                                         self->priv->name,
-                                         g_udev_device_get_subsystem (port),
-                                         g_udev_device_get_name (port));
-        g_simple_async_result_complete_in_idle (async_result);
-        goto out;
+        g_task_return_new_error (task,
+                                 MM_CORE_ERROR,
+                                 MM_CORE_ERROR_FAILED,
+                                 "(%s) Missing port probe for port (%s/%s)",
+                                 self->priv->name,
+                                 g_udev_device_get_subsystem (port),
+                                 g_udev_device_get_name (port));
+        g_object_unref (task);
+        return;
     }
 
     /* Before launching any probing, check if the port is a net device. */
     if (g_str_equal (g_udev_device_get_subsystem (port), "net")) {
         mm_dbg ("(%s) [%s] probing deferred until result suggested",
-                self->priv->name,
-                g_udev_device_get_name (port));
-        g_simple_async_result_set_op_res_gpointer (
-            async_result,
-            GUINT_TO_POINTER (MM_PLUGIN_SUPPORTS_PORT_DEFER_UNTIL_SUGGESTED),
-            NULL);
-        g_simple_async_result_complete_in_idle (async_result);
-        goto out;
+                self->priv->name, g_udev_device_get_name (port));
+        g_task_return_int (task, MM_PLUGIN_SUPPORTS_PORT_DEFER_UNTIL_SUGGESTED);
+        g_object_unref (task);
+        return;
     }
 
     /* Build flags depending on what probing needed */
+    probe_run_flags = MM_PORT_PROBE_NONE;
     if (!g_str_has_prefix (g_udev_device_get_name (port), "cdc-wdm")) {
         /* Serial ports... */
-        probe_run_flags = MM_PORT_PROBE_NONE;
         if (self->priv->at)
             probe_run_flags |= MM_PORT_PROBE_AT;
         else if (self->priv->single_at)
@@ -707,7 +745,6 @@ mm_plugin_supports_port (MMPlugin *self,
             probe_run_flags |= MM_PORT_PROBE_QCDM;
     } else {
         /* cdc-wdm ports... */
-        probe_run_flags = MM_PORT_PROBE_NONE;
         if (self->priv->qmi && !g_strcmp0 (mm_device_utils_get_port_driver (port), "qmi_wwan"))
             probe_run_flags |= MM_PORT_PROBE_QMI;
         else if (self->priv->mbim && !g_strcmp0 (mm_device_utils_get_port_driver (port), "cdc_mbim"))
@@ -729,11 +766,9 @@ mm_plugin_supports_port (MMPlugin *self,
     /* If no explicit probing was required, just request to grab it without probing anything.
      * This may happen, e.g. with cdc-wdm ports which do not need QMI/MBIM probing. */
     if (probe_run_flags == MM_PORT_PROBE_NONE) {
-        g_simple_async_result_set_op_res_gpointer (async_result,
-                                                   GUINT_TO_POINTER (MM_PLUGIN_SUPPORTS_PORT_DEFER_UNTIL_SUGGESTED),
-                                                   NULL);
-        g_simple_async_result_complete_in_idle (async_result);
-        goto out;
+        g_task_return_int (task, MM_PLUGIN_SUPPORTS_PORT_DEFER_UNTIL_SUGGESTED);
+        g_object_unref (task);
+        return;
     }
 
     /* If a modem is already available and the plugin says that only one AT port is
@@ -753,11 +788,13 @@ mm_plugin_supports_port (MMPlugin *self,
     }
 
     /* Setup async call context */
-    ctx = g_new (PortProbeRunContext, 1);
-    ctx->self = g_object_ref (self);
+    ctx = g_slice_new0 (PortProbeRunContext);
+    ctx->self   = g_object_ref (self);
     ctx->device = g_object_ref (device);
-    ctx->result = g_object_ref (async_result);
-    ctx->flags = probe_run_flags;
+    ctx->flags  = probe_run_flags;
+
+    /* Store context in task */
+    g_task_set_task_data (task, ctx, (GDestroyNotify) port_probe_run_context_free);
 
     /* Launch the probe */
     probe_list_str = mm_port_probe_flag_build_string_from_mask (ctx->flags);
@@ -774,13 +811,9 @@ mm_plugin_supports_port (MMPlugin *self,
                        self->priv->send_lf,
                        self->priv->custom_at_probe,
                        self->priv->custom_init,
-                       (GAsyncReadyCallback)port_probe_run_ready,
-                       ctx);
-
-out:
-    g_object_unref (async_result);
-    if (probe)
-        g_object_unref (probe);
+                       cancellable,
+                       (GAsyncReadyCallback) port_probe_run_ready,
+                       task);
 }
 
 /*****************************************************************************/
@@ -865,22 +898,43 @@ mm_plugin_create_modem (MMPlugin  *self,
                                            "unsupported subsystem: '%s'",
                                            mm_port_probe_get_port_subsys (probe));
             }
+            /* Ports that are explicitly blacklisted will be grabbed as ignored */
+            else if (mm_port_probe_is_ignored (probe)) {
+                mm_dbg ("(%s/%s): port is blacklisted",
+                        mm_port_probe_get_port_subsys (probe),
+                        mm_port_probe_get_port_name (probe));
+                grabbed = mm_base_modem_grab_port (modem,
+                                                   mm_port_probe_get_port_subsys (probe),
+                                                   mm_port_probe_get_port_name (probe),
+                                                   mm_port_probe_get_parent_path (probe),
+                                                   MM_PORT_TYPE_IGNORED,
+                                                   MM_PORT_SERIAL_AT_FLAG_NONE,
+                                                   &inner_error);
+            }
 #if !defined WITH_QMI
             else if (mm_port_probe_get_port_type (probe) == MM_PORT_TYPE_NET &&
                      !g_strcmp0 (mm_device_utils_get_port_driver (mm_port_probe_peek_port (probe)), "qmi_wwan")) {
-                grabbed = FALSE;
-                inner_error = g_error_new (MM_CORE_ERROR,
-                                           MM_CORE_ERROR_UNSUPPORTED,
-                                           "ignoring QMI net port");
+                /* Try to generically grab the port, but flagged as ignored */
+                grabbed = mm_base_modem_grab_port (modem,
+                                                   mm_port_probe_get_port_subsys (probe),
+                                                   mm_port_probe_get_port_name (probe),
+                                                   mm_port_probe_get_parent_path (probe),
+                                                   MM_PORT_TYPE_IGNORED,
+                                                   MM_PORT_SERIAL_AT_FLAG_NONE,
+                                                   &inner_error);
             }
 #endif
 #if !defined WITH_MBIM
             else if (mm_port_probe_get_port_type (probe) == MM_PORT_TYPE_NET &&
                      !g_strcmp0 (mm_device_utils_get_port_driver (mm_port_probe_peek_port (probe)), "cdc_mbim")) {
-                grabbed = FALSE;
-                inner_error = g_error_new (MM_CORE_ERROR,
-                                           MM_CORE_ERROR_UNSUPPORTED,
-                                           "ignoring MBIM net port");
+                /* Try to generically grab the port, but flagged as ignored */
+                grabbed = mm_base_modem_grab_port (modem,
+                                                   mm_port_probe_get_port_subsys (probe),
+                                                   mm_port_probe_get_port_name (probe),
+                                                   mm_port_probe_get_parent_path (probe),
+                                                   MM_PORT_TYPE_IGNORED,
+                                                   MM_PORT_SERIAL_AT_FLAG_NONE,
+                                                   &inner_error);
             }
 #endif
             else if (MM_PLUGIN_GET_CLASS (self)->grab_port)
